#include "circuit.h"

#include <queue>
#include <unordered_map>
#include <unordered_set>
#include "assert.h"

void Circuit::init_topo_index() {
    int topo = 1;
    std::queue<Gate*> q;

    std::unordered_map<Gate*, int> ins;
    for(Gate* gate : gates) {
        ins[gate] = gate->fanins.size();
    }

    for(auto in : PIs) {
        in->id = topo++;
        q.push(in);
    }

    while(!q.empty()) {
        Gate* g = q.front(); q.pop();
        for(Gate* out : g->fanouts) {
            ins[out]--;
            if(ins[out] == 0) {
                out->id = topo++;
                q.push(out);
            }
        }
    }

    // 计算反向拓扑序
    topo = 1;
    std::unordered_map<Gate*, int> outs;

    for(auto out : POs) {
        out->rtopo = topo++;
        q.push(out);
    }

    while(!q.empty()) {
        Gate* g = q.front(); q.pop();
        rtopo_gates.push_back(g);
        for(Gate* in : g->fanins) {
            outs[in]++;
            if(outs[in] == in->fanouts.size()) {
                in->rtopo = topo++;
                q.push(in);
            }
        }
    }
}

void Circuit::init_gate_level() {
    MAX_GATE_LEVEL = 0;
    std::queue<Gate*> q;

    for(Gate* gate : gates) {
        gate->level = -1;
    }

    for(auto pi: PIs) {
        pi->level = 0;
        q.push(pi);
    }

    while(!q.empty()) {
        Gate* g = q.front(); q.pop();

        MAX_GATE_LEVEL = std::max(MAX_GATE_LEVEL, g->level);

        for(Gate* out : g->fanouts) {
            if(out->level == -1) {
                out->level = g->level + 1;
                q.push(out);
            }
        }
    }

    for(Gate* g : gates) {
        assert(g->level != -1);
    }
}

void Circuit::print_gates() {
    static const char* type2name[9] = {"AND", "NAND", "OR", "NOR", "XOR", "XNOR", "NOT", "BUF", "IN"};
    for(Gate* gate : gates) {
        printf("Gate: %3s (t:%4s v:%d pi:%d po:%d s0:%d s1:%d fpl0:%d fpl1:%d)  Inputs:", gate->name.c_str(), type2name[gate->type], gate->value, gate->isPI, gate->isPO, gate->fault_detected[0], gate->fault_detected[1], gate->fault_propagated_len[0], gate->fault_propagated_len[1]);
        for(Gate* in : gate->fanins) {
            printf(" %s(%d)", in->name.c_str(), gate->is_detected(in));
        }
        printf("\n");
    }
}


void Circuit::init_reigons() {
    for(Gate* g : gates) {
        for(Gate* in : g->fanins) {
            g->reigon.push_back(in);
        }
        for(Gate* out : g->fanouts) {
            for(Gate* in : out->fanins) {
                if(in == g) continue;
                g->reigon.push_back(in);
            }
        }
        std::sort(g->reigon.begin(), g->reigon.end());
        g->reigon.erase(unique(g->reigon.begin(), g->reigon.end()), g->reigon.end());
    }

    std::unordered_set<Gate*> r_set;

    // score reigon
    for(Gate* gate : gates) {
        r_set.clear();    

        for(Gate* out : gate->fanouts) {
            r_set.insert(out);
            for(Gate* in : out->fanins) {
                r_set.insert(in);
                for(Gate *out : in->fanouts) {
                    r_set.insert(out);
                    for(Gate *in : out->fanins) {
                        r_set.insert(in);
                    }
                }
            }
        }

        for(Gate* in : gate->fanins) {
            r_set.insert(in);
            for(Gate* out : in->fanouts) {
                r_set.insert(out);
                for(Gate* in : out->fanins) {
                    r_set.insert(in);
                }
            }
        }
        
        r_set.erase(gate);

        for(Gate* g : r_set) {
            gate->score_reigon.push_back(g);
        }
    }
}

// bool Circuit::is_valid_circuit() {

    // ll flip_total_weight = 0;
    // ll stem_total_weight = 0;
    // ll fault_total_weight = 0;

    // int flip_total_cnt = 0;
    // int stem_total_cnt = 0;
    // int fault_total_cnt = 0;

    // ll fault_propagate_score = 0;

    // //printf("flip: %d, stem: %d, fault:%d\n", flip_total_weight, stem_total_weight, fault_total_weight);

    // for(Gate* g : gates) {

    //     fault_propagate_score += g->fault_propagate_len[0] * fault_weight[g->id][0];
    //     fault_propagate_score += g->fault_propagate_len[1] * fault_weight[g->id][1];

    //     if(flip_need_update[g->id]) {
    //         flip_total_weight += flip_weight[g->id];
    //         flip_total_cnt++;
    //     }

    //     if(g->stem && g->cal_value() != g->value) {
    //         stem_total_weight += stem_weight[g->id];
    //     }

    //     if(g->stem && g->cal_value() == g->value) {
    //         stem_total_cnt++;
    //     }

    //     if(g->cal_propagate_len(0) != g->fault_propagate_len[0] || g->cal_propagate_len(1) != g->fault_propagate_len[1]) {
    //         printf("WRONG-PRO-LEN: %s \n", g->name.c_str());
    //         print_gates();
    //         return false;
    //     }

    //     if(g->sa[0]) {
    //         fault_total_weight += fault_weight[g->id][0];
    //         fault_total_cnt += 1;
    //     }

    //     if(g->sa[1]) {
    //         fault_total_weight += fault_weight[g->id][1];
    //         fault_total_cnt += 1;
    //     }

    //     if(g->stem) {
    //         assert(stem_satisfied[g->id] == (g->cal_value() == g->value));
    //     }

    //     // 检查门的赋值情况
    //     if(g->cal_value() != g->value) {
    //         printf("WRONG-ASSGIN: %s \n", g->name.c_str());
    //         return false;
    //     }
        
    //     // 检查 PO 的传播设定是否正确
    //     if(g->isPO) {
    //         if(g->sa[g->value] != 0 || g->sa[!g->value] == 0 ) {
    //             printf("WRONG-PO: %s \n", g->name.c_str());
    //         }
    //         continue;
    //     }

    //     // 非 PO 情况下检查故障传播是否正确

    //     bool sa0 = false;
    //     bool sa1 = false;

    //     for(Gate* out : g->fan_outs) {
    //         if(out->cal_value() != out->value) {
    //             assert(out->stem);
    //             continue;
    //         }

    //         g->value = !g->value;

    //         if(out->cal_value() != out->value) {
    //             sa0 |= out->is_propagated() && !g->value;
    //             sa1 |= out->is_propagated() && g->value;
    //         }

    //         g->value = !g->value;
    //     }

    //     if(sa0 != g->sa[0] || sa1 != g->sa[1]) {
    //         printf("WRONG-SA: %s \n", g->name.c_str());
    //         return false;
    //     }
    // }

    // if(this->flip_total_weight != flip_total_weight || this->stem_total_weight != stem_total_weight || this->fault_total_weight != fault_total_weight) {
    //     printf("CIRCUIT CHECK FAILED!\n");
    //     printf("[wrong] flip: %d, stem: %d, fault:%d\n", this->flip_total_weight, this->stem_total_weight, this->fault_total_weight);
    //     printf("[right] flip: %d, stem: %d, fault:%d\n", flip_total_weight, stem_total_weight, fault_total_weight);
    //     return false;
    // }

    // if(this->flip_total_cnt != flip_total_cnt || this->stem_total_cnt != stem_total_cnt || this->fault_total_cnt != fault_total_cnt) {
    //     printf("CIRCUIT CHECK FAILED!\n");
    //     printf("[wrong] flip_cnt: %d, stem_cnt: %d, fault_cnt:%d\n", this->flip_total_cnt, this->stem_total_cnt, this->fault_total_cnt);
    //     printf("[right] flip_cnt: %d, stem_cnt: %d, fault_cnt:%d\n", flip_total_cnt, stem_total_cnt, fault_total_weight);
    //     return false;
    // }

    // printf("%lld %lld\n", fault_propagate_score , this->fault_propagate_score);
    // assert(fault_propagate_score == this->fault_propagate_score);

//     return true;
// }


